'******************************************************************************
' Copyright (c) 2012, Dave Hein
' See end of file for terms of use.
'******************************************************************************
OBJ
  sys : "sysdefs"

CON
  _clkmode = xinput
  _clkfreq = 0

  ' SPI mail box
  pSPI_engine_cog     = sys#SPI_engine_cog
  pSPI_command        = sys#SPI_command
  pSPI_block_index    = sys#SPI_block_index
  pSPI_buffer_address = sys#SPI_buffer_address

  ' Start address of the LMM C kernel in hub RAM
  START_ADDR     = $20

  ' LMM Kernel Registers
  r1 = 1
  r2 = 2
  r5 = 5
  lr = 15
  pc = 17

DAT
' This is used to replace an instruction in the LMM kernel so that we can
' use a stack address other than $8000.  The source address for __C_LOCK_PTR
' will be copied from the instruction at cog location 5.
patch   rdlong r1, 0 wz

' This exit routine is used to terminate stand-alone programs by rebooting
exit    mov    r1, #128
        clkset r1

PUB LoadFile(argc, argv, argsize, sector, size) | argv1, cogspi, i

  ' The locks have been freed by the caller, so allocate lock 0 so the PAR
  ' patch works for threads.  It depends on a non-zero lock number.
  locknew

  ' Get the SPI cog number
  cogspi := long[pSPI_engine_cog] - 1

  ' Copy the arg list to just before this object
  argv1 := sys#start - argsize
  bytemove(argv1, argv, argsize)

  ' Fix argv pointer
  repeat i from 0 to argc - 1
    long[argv1][i] += argv1 - argv

  ' Read the program into memory
  repeat (size + 511) >> 9
    long[pSPI_block_index] := sector++
    long[pSPI_buffer_address] := result
    long[pSPI_command] := "r"
    repeat while long[pSPI_command] == "r"
    if long[pSPI_command] < 0
      reboot
    result += 512

  ' Release the SD card
  long[pSPI_command] := "z"
  repeat while long[pSPI_command] == "z"

  ' Stop the SPI cog
  cogstop(cogspi)

  ' Patch cog location $2 to test if lock has been allocated instead
  ' of testing for stack address in PAR is not equal to $8000
  long[START_ADDR][r2] := patch | (long[START_ADDR][r5] & $1ff)

  ' Patch the address of the arg list into the app
  ' The pc contains the entry address, and the third long after
  ' the entry address points to the arg list
  result := long[START_ADDR][pc]
  long[result][3] := argv1

  ' Patch in the address of the exit code
  ' The lr register contains the address of the exit routine, and
  ' the fist word after the start of the exit routine is the address
  ' of the exit hook routine.
  result := long[START_ADDR][lr]
  long[result][1] := @exit

  ' Start app in cog 0
  coginit(0, START_ADDR, argv1)

  ' Stop this cog if we are still runnning
  cogstop(cogid)

{{
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
}}

